version(build) { pragma(link, wsock32); }

private import std.socket;
private import std.thread;
private import std.stream;

version(Windows)
{
	private import std.c.windows.windows;
	private import std.c.windows.winsock;
}
else
{
	private import std.c.linux.linux;
	private import std.c.linux.socket;
}

import util;

class AsyncSocket : Stream
{
	this(size_t bufferSize = 8192)
	{
		buffer = new ubyte[bufferSize];
		updateAttribs();
	}
	
	void disconnect()
	{
		if (!s) return ;
		s.close();
		s = null;
		updateAttribs();
	}
	
	void connect(char[] host, short port)
	{
		disconnect();
		s = new TcpSocket(new InternetAddress(host,port));
		updateAttribs();
	}
	
	bool isConnected()
	{
		return isopen;
	}
	
	void stop()
	{
		if (!t) return;
		terminate = true;
		t.wait();
		t = null;
	}
	
	void start()
	{
		stop();
		terminate = false;
		t = new Thread(&run);
		t.start();
	}	
	
	bool isRunning()
	{
		return !terminate;
	}
	
	size_t readBlock(void* bptr, size_t size)
	{
		assertReadable();
		
		int r;	
		
		synchronized(this)
		{
			r = bytes;
			if (r > size) r = size;
			memcpy(bptr,buffer,r);
			bytes -= r;
			if (bytes) memcpy(buffer,&buffer[r],bytes);
		}
		
		return r;
	}
	
	size_t writeBlock(void* buffer, size_t size)
	{
		assertWriteable();
		return s.send(buffer[0..size]);
	}
	
	ulong seek(long offset, SeekPos whence)
	{
		assertSeekable();
		//never seekable
	}
	
private:
	bool terminate = true;
	Thread t;
	
	TcpSocket s;
	
	bool lockbuffer = false;
	ubyte[] buffer;
	size_t bytes;
	
	int run()
	{
		size_t space;
		ubyte[1] b;		
		int r;
		
		while(!terminate)
		{
			synchronized(this) space = buffer.length - bytes;
			if (space == 0)
			{
				msleep(100);
				continue;
			}
			
			r = s.receive(b);
			if (r == 0)
			{
				synchronized(this) isopen = false;				
				break;
			}
			
			if (r == SOCKET_ERROR)
			{
				if (!wouldBlock()) break;
				msleep(100);
				continue;
			}
			
			synchronized(this) buffer[bytes++] = b[0];
		}
		
		synchronized(this) terminate = true;
		
		return 0;
	}
	
	void updateAttribs()
	{
		if (s)
		{
			readable = writeable = true;
			seekable = false;
      			isopen = true;
		}
		else
		{
			readable = writeable = seekable = false;			
      			isopen = false;
		}
	}
	
	bool wouldBlock()
	{
		version(Windows) return WSAGetLastError() == WSAEWOULDBLOCK;
		else return errno == EWOULDBLOCK;
	}
}
